# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['src',
 'src.intelligenceArtificielle',
 'src.intelligenceArtificielle.structureDonnees']

package_data = \
{'': ['*']}

modules = \
['sunrise']
install_requires = \
['keras>=2.3.1,<3.0.0',
 'matplotlib>=3.2.1,<4.0.0',
 'numpy>=1.18.3,<2.0.0',
 'scipy>=1.4.1,<2.0.0',
 'tensorflow>=2.2.0,<3.0.0']

entry_points = \
{'console_scripts': ['sunrise = sunrise:lancement_interface_graphique']}

setup_kwargs = {
    'name': 'sunriseces',
    'version': '0.0.3',
    'description': "Outils d'inférence de paramètres pour thermique de bâtiments",
    'long_description': '[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n![Python application](https://github.com/18tbr/sunrise-refonte/workflows/Python%20application/badge.svg)\n[![License](https://img.shields.io/github/license/18tbr/sunrise-refonte)](https://github.com/18tbr/sunrise-refonte/blob/master/LICENSE)\n![Windows Support](https://img.shields.io/badge/Windows-Support-brightgreen.svg)\n![Linux Support](https://img.shields.io/badge/Linux-Support-brightgreen.svg)\n![Mines de Paris](https://img.shields.io/badge/Organisme-Mines%20de%20Paris-blue)\n# Mines de Paris - CES\nLe dépôt GitHub pour le projet SunRise de 2020.\n\n<p align="center">\n    <a href="https://sites.google.com/view/2019-2020-gr12/home" target="_blank">\n        <img src="https://github.com/18tbr/sunrise-refonte/blob/master/assets/logoProjet2020.png?raw=true" height="250px">\n    </a>\n</p>\n\n## Le projet SunRise\nLe projet SunRise est un projet lancé en 2017 par le [CES](http://www.ces.mines-paristech.fr/Accueil/) (Centre d\'Efficacité des Systèmes) des [Mines de Paris](http://www.mines-paristech.fr/). Ce projet a pour but de __valider des modèles novateurs de thermique :fire: de bâtiment :house:__ du CES à l\'aide d\'une expérience en chambre climatique.\n\nUne particularité de l\'expérience du CES (qui donne son nom au projet) est qu\'elle utilise un __soleil :sunny: artificiel mobile de grande dimension__ (~2m de diamètre) afin de bénéficier de davantage de flexibilité lors de l\'évaluation de l\'influence des rayons lumineux :flashlight:\n\nPour plus de détails, vous pouvez visiter le [site du projet](https://sites.google.com/view/2019-2020-gr12/home) :point_left: ou même la [vidéo de présentation du projet](https://youtu.be/zrpTuERl-vk) :movie_camera:\n\n## Ce dépôt\nCe dépôt contient le travail d\'une équipe d\'étudiants des Mines de Paris sur une __méthode d\'identification de paramètres d\'intérêt__ :monocle_face: à partir de mesures de thermique de bâtiment. Il s\'agit d\'un élément préalable à la réalisation de l\'expérience SunRise.\n\nLa méthode développée ici est inhabituelle, et l\'algorithme proposé tente de fournir une solution à l\'inférence de paramètres en utilisant des __méthodes de Machine Learning__ :robot:\n\n## Package Python\n\nCe projet utilise l\'outil [Poetry](https://python-poetry.org) pour gérer les dépendances et faire un package pip utilisable plus facilement. Le package pip du projet se nomme [sunriseces](https://pypi.org/project/sunriseces/).\n\nPour installer le package, vous pouvez donc utiliser :\n\n```sh\n# Si pip se trouve dans votre PATH\npip install sunriseces\n\n# Sinon sous Linux\npython3 -m pip install sunriseces\n\n# De même sous Windows\npython -m pip install sunriseces\n```\n\nUne fois le package installé, vous pourrez utiliser l\'outil graphique en utilisant l\'une des commandes :\n\n```sh\n# Si le dossier des modules exécutables python se trouve dans votre PATH\nsunrise\n\n# Sinon sous Linux\npython3 -m sunrise\n\n# De même sous Windows\npython -m sunrise\n```\n\n## Contribution\nSi vous __souhaitez contribuer à ce dépôt :heart:__, nous vous invitons à consulter les différents documents du [dossier docs](https://github.com/18tbr/sunrise-refonte/tree/master/docs) qui détaillent les façons de contribuer au projet, les conventions de nommage, etc...\n\n## License\nLe code de ce dépôt est _sauf contre-indication_ mis à disposition sous license [GPLv3](https://www.gnu.org/licenses/gpl-3.0.en.html).\n\n<p align="center">\n  <a href="http://www.mines-paristech.fr/" target="_blank">\n      <img src="https://github.com/18tbr/sunrise-refonte/blob/master/assets/logoMinesParisTech.png?raw=true" height="100px">\n  </a>\n</p>\n\n<p align="center">\n  <a href="http://www.ces.mines-paristech.fr/Accueil/" target="_blank">\n      <img src="https://github.com/18tbr/sunrise-refonte/blob/master/assets/logoCES.png?raw=true" height="75px">\n  </a>\n</p>\n',
    'author': 'Voir GitHub',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://sites.google.com/view/2019-2020-gr12/home',
    'packages': packages,
    'package_data': package_data,
    'py_modules': modules,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
